@@ -183,7 +183,7 @@ void Application::createWindow(const Profile::Ptr &profile, const QString &direc
 
 void Application::detachTerminals(ViewSplitter *splitter,const QHash<TerminalDisplay*, Session*>& sessionsMap)
 {
-    MainWindow *currentWindow = qobject_cast<MainWindow*>(sender());
+    auto *currentWindow = qobject_cast<MainWindow*>(sender());
     MainWindow *window = newMainWindow();
     ViewManager *manager = window->viewManager();
 
@@ -303,7 +303,11 @@ bool Application::processTabsFromFileArgs(MainWindow *window)
         }
 
         QHash<QString, QString> lineTokens;
+#if (QT_VERSION >= QT_VERSION_CHECK(5, 15, 0))
+        QStringList lineParts = lineString.split(QStringLiteral(";;"), Qt::SkipEmptyParts);
+#else
         QStringList lineParts = lineString.split(QStringLiteral(";;"), QString::SkipEmptyParts);
+#endif
 
         for (int i = 0; i < lineParts.size(); ++i) {
             QString key = lineParts.at(i).section(QLatin1Char(':'), 0, 0).trimmed().toLower();
@@ -336,6 +340,7 @@ void Application::createTabFromArgs(MainWindow *window, const QHash<QString, QSt
     const QString &command = tokens[QStringLiteral("command")];
     const QString &profile = tokens[QStringLiteral("profile")];
     const QString &workdir = tokens[QStringLiteral("workdir")];
+    const QColor &color = tokens[QStringLiteral("tabcolor")];
 
     Profile::Ptr baseProfile;
     if (!profile.isEmpty()) {
@@ -365,6 +370,12 @@ void Application::createTabFromArgs(MainWindow *window, const QHash<QString, QSt
         shouldUseNewProfile = true;
     }
 
+    // For tab color support
+    if (color.isValid()) {
+        newProfile->setProperty(Profile::TabColor, color);
+        shouldUseNewProfile = true;
+    }
+
     if (m_parser->isSet(QStringLiteral("workdir"))) {
         newProfile->setProperty(Profile::Directory, m_parser->value(QStringLiteral("workdir")));
         shouldUseNewProfile = true;
@@ -543,9 +554,8 @@ Profile::Ptr Application::processProfileChangeArgs(Profile::Ptr baseProfile)
 
     if (shouldUseNewProfile) {
         return newProfile;
-    } else {
-        return baseProfile;
     }
+    return baseProfile;
 }
 
 void Application::startBackgroundMode(MainWindow *window)
