@@ -117,6 +117,8 @@ public:
 
     void setScrollFullPage(bool fullPage);
     bool scrollFullPage() const;
+    void setHighlightScrolledLines(bool highlight);
+
     /**
      * Returns the display's filter chain.  When the image for the display is updated,
      * the text is passed through each filter in the chain.  Each filter can define
@@ -371,6 +373,11 @@ public:
     /** See setAlternateScrolling() */
     bool alternateScrolling() const;
 
+    bool hasCompositeFocus() const
+    {
+        return _hasCompositeFocus;
+    }
+
 public Q_SLOTS:
     /**
      * Scrolls current ScreenWindow
@@ -546,8 +553,7 @@ Q_SIGNALS:
 
     void sendStringToEmu(const QByteArray &local8BitString);
 
-    void focusLost();
-    void focusGained();
+    void compositeFocusChanged(bool focused);
 
 protected:
     // events
@@ -610,6 +616,7 @@ protected Q_SLOTS:
     void scrollBarPositionChanged(int value);
     void blinkTextEvent();
     void blinkCursorEvent();
+    void highlightScrolledLinesEvent();
 
 private Q_SLOTS:
 
@@ -630,6 +637,11 @@ private:
     void drawContents(QPainter &painter, const QRect &rect);
     // draw a transparent rectangle over the line of the current match
     void drawCurrentResultRect(QPainter &painter);
+    // draw a thin highlight on the left of the screen for lines that have been scrolled into view
+    void highlightScrolledLines(QPainter& painter);
+    // compute which region need to be repainted for scrolled lines highlight
+    QRect highlightScrolledLinesRegion(void);
+
     // draws a section of text, all the text in this section
     // has a common color and style
     void drawTextFragment(QPainter &painter, const QRect &rect, const QString &text,
@@ -674,6 +686,10 @@ private:
     // current size in columns and lines
     void showResizeNotification();
 
+    // applies changes to scrollbarLocation to the scroll bar and
+    // if @propagate is true, propagates size information
+    void applyScrollBarPosition(bool propagate = true);
+
     // scrolls the image by a number of lines.
     // 'lines' may be positive ( to scroll the image down )
     // or negative ( to scroll the image up )
@@ -689,6 +705,8 @@ private:
 
     void paintFilters(QPainter &painter);
 
+    void setupHeaderVisibility();
+
     // returns a region covering all of the areas of the widget which contain
     // a hotspot
     QRegion hotSpotRegion() const;
@@ -880,6 +898,18 @@ private:
 
     bool _drawOverlay;
     Qt::Edge _overlayEdge;
+
+    struct {
+        bool enabled = false;
+        QRect rect;
+        int previousScrollCount = 0;
+        QTimer *timer = nullptr;
+    } _highlightScrolledLinesControl;
+    static const int HIGHLIGHT_SCROLLED_LINES_WIDTH = 3;
+
+    bool _hasCompositeFocus;
+
+    QSharedPointer<Filter::HotSpot> _currentlyHoveredHotspot;
 };
 
 class AutoScrollHandler : public QObject
@@ -899,6 +929,29 @@ private:
 
     int _timerId;
 };
+
+// Watches compositeWidget and all its focusable children,
+// and emits focusChanged() signal when either compositeWidget's
+// or a child's focus changed.
+// Limitation: children added after the object was created
+// will not be registered.
+class CompositeWidgetFocusWatcher : public QObject
+{
+    Q_OBJECT
+
+public:
+    explicit CompositeWidgetFocusWatcher(QWidget *compositeWidget, QObject *parent);
+    bool eventFilter(QObject *watched, QEvent *event) Q_DECL_OVERRIDE;
+
+Q_SIGNALS:
+    void compositeFocusChanged(bool focused);
+
+private:
+    void registerWidgetAndChildren(QWidget *widget);
+
+    QWidget *_compositeWidget;
+};
+
 }
 
 #endif // TERMINALDISPLAY_H
